from shortcrust.gl2 import *
from shortcrust.matrix import vec3, mat4
from shortcrust.buffer import AttributeBuffer
from shortcrust.geometry import Sphere

from shaders.phong import PhongSolidColorShader
from shaders.texture import TextureShader

from scenes.robots import Robot, Robokid
import math


class SkyMesh(object):
	mode = GL_TRIANGLE_STRIP

	def __init__(self, app):
		self.positions = AttributeBuffer([
			(-60, -60, 60), (60, -60, 60), (-60, 60, 60), (60, 60, 60)
		])
		self.vertex_count = self.positions.element_count

		self.texture_positions = AttributeBuffer([
			(0.0, 0.0), (1.0, 0.0), (0.0, 1.0), (1.0, 1.0)
		])

		self.texture = app.textures['sunset']

	def draw(self):
		glDrawArrays(self.mode, 0, self.vertex_count)


class FaderLayer(object):
	def __init__(self, shader, mesh):
		self.shader = shader
		self.mesh = mesh

	def draw(self, beat):
		if beat < 36:
			alpha = 1 - (beat - 34) / 2
		else:
			return
		self.shader.use()
		self.shader.draw_mesh(self.mesh, alpha)


class HorizonScene(object):
	def __init__(self, app):
		# glClearColor(0.588, 0.659, 0.706, 1.0)
		glEnable(GL_DEPTH_TEST)

		self.app = app
		self.fader_layer = FaderLayer(app.fader_shader, app.fader_mesh)

		self.phong_shader = PhongSolidColorShader()
		self.texture_shader = TextureShader()

		self.mvMatrix = mat4.create()
		self.ground_mvMatrix = mat4.create()
		self.pMatrix = mat4.create()
		self.lightLocation = vec3.create([-3, 30, 0])
		self.transformedLightLocation = vec3.create()

		self.phong_shader.use()
		self.phong_shader.pMatrix = mat4.perspective(45, self.app.aspect_ratio, 0.1, 100.0)

		self.texture_shader.use()
		self.texture_shader.pMatrix = mat4.perspective(45, self.app.aspect_ratio, 0.1, 100.0)

		self.robot = Robot(self.phong_shader)
		self.robokid = Robokid(self.phong_shader)

		self.ground_mesh = Sphere(r=20, min_lat_deg=0, max_lat_deg=90, min_lng_deg=135, max_lng_deg=135+90).to_mesh(
			texture=app.textures['ground']
		)
		self.sky_mesh = SkyMesh(app)

	def draw(self, beat):
		self.robot.animate(beat)
		self.robokid.animate(beat)

		glViewport(0, 0, self.app.width, self.app.height)
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

		mat4.lookAt(
			[0, 2, -6],  # eye
			[0, 1.5, 0],  # lookat
			[0, 1, 0],  # up
			self.mvMatrix)

		self.texture_shader.use()
		self.texture_shader.mvMatrix = self.mvMatrix

		self.texture_shader.draw_mesh(self.sky_mesh)

		mat4.translate(self.mvMatrix, [0, -20, 0], self.ground_mvMatrix)
		mat4.rotateX(self.ground_mvMatrix, math.pi / 4)
		self.texture_shader.mvMatrix = self.ground_mvMatrix
		self.texture_shader.draw_mesh(self.ground_mesh)

		self.phong_shader.use()

		mat4.multiplyVec3(self.mvMatrix, self.lightLocation, self.transformedLightLocation)
		self.phong_shader.lightLocation = self.transformedLightLocation

		start_beat = beat - 34

		mat4.translate(self.mvMatrix, [0, -20, 0])
		mat4.rotateX(self.mvMatrix, math.pi/4-start_beat / 24)
		mat4.translate(self.mvMatrix, [0, 20, 0])

		mat4.translate(self.mvMatrix, [-1.5, 0, 0])
		self.robot.draw(self.mvMatrix)
		mat4.translate(self.mvMatrix, [1.5, 0, 0])
		self.robot.draw(self.mvMatrix)
		mat4.translate(self.mvMatrix, [1.5, 0, 0])
		self.robot.draw(self.mvMatrix)
		mat4.translate(self.mvMatrix, [-2.25, 0, 1])
		self.robokid.draw(self.mvMatrix)

		self.fader_layer.draw(beat)

		# sleep(0.01)
