from shortcrust.gl2 import *
from shortcrust.matrix import vec3, mat4
from shortcrust.buffer import AttributeBuffer

from shaders.phong import PhongSolidColorShader
from shaders.texture import TextureShader

from scenes.robots import Robot, Robokid
from scenes.fader import BlackFaderShader


class GroundMesh(object):
	mode = GL_TRIANGLE_STRIP

	def __init__(self, app):
		self.positions = AttributeBuffer([
			(-20, 0, -10), (15, 0, -10), (-20, 0, 30), (15, 0, 30)
		])
		self.vertex_count = self.positions.element_count

		self.texture_positions = AttributeBuffer([
			(0.0, 0.0), (1.0, 0.0), (0.0, 1.0), (1.0, 1.0)
		])

		self.texture = app.textures['ground']

	def draw(self):
		glDrawArrays(self.mode, 0, self.vertex_count)


class SkyMesh(object):
	mode = GL_TRIANGLE_STRIP

	def __init__(self, app):
		self.positions = AttributeBuffer([
			(-20, 0, 10), (15, 0, 10), (-20, 15, 10), (15, 15, 10)
		])
		self.vertex_count = self.positions.element_count

		self.texture_positions = AttributeBuffer([
			(0.0, 0.0), (1.0, 0.0), (0.0, 1.0), (1.0, 1.0)
		])

		self.texture = app.textures['sunset']

	def draw(self):
		glDrawArrays(self.mode, 0, self.vertex_count)


class FaderLayer(object):
	def __init__(self, shader, mesh):
		self.shader = shader
		self.mesh = mesh

	def draw(self, beat):
		if beat < 75:
			return
		else:
			alpha = (beat - 75) / 3
		self.shader.use()
		self.shader.draw_mesh(self.mesh, alpha)


class PanScene(object):
	def __init__(self, app):
		# glClearColor(0.588, 0.659, 0.706, 1.0)
		glEnable(GL_DEPTH_TEST)

		self.app = app
		self.fader_layer = FaderLayer(BlackFaderShader(), app.fader_mesh)

		self.phong_shader = PhongSolidColorShader()
		self.texture_shader = TextureShader()

		self.mvMatrix = mat4.create()
		self.pMatrix = mat4.create()
		self.lightLocation = vec3.create([-3, 30, 0])
		self.transformedLightLocation = vec3.create()

		self.phong_shader.use()
		self.phong_shader.pMatrix = mat4.perspective(45, self.app.aspect_ratio, 0.1, 100.0)

		self.texture_shader.use()
		self.texture_shader.pMatrix = mat4.perspective(45, self.app.aspect_ratio, 0.1, 100.0)

		self.robot = Robot(self.phong_shader)
		self.robokid = Robokid(self.phong_shader)

		self.ground_mesh = GroundMesh(app)
		self.sky_mesh = SkyMesh(app)

	def draw(self, beat):
		self.robot.animate(beat)
		self.robokid.animate(beat)

		glViewport(0, 0, self.app.width, self.app.height)
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

		movement = (beat-56) * 0.3

		mat4.lookAt(
			[2.2 - (beat-56) / 6.0, 2.5, -5 - movement],  # eye
			[2.2 - (beat-56) / 6.0, 2.0, -movement],  # lookat
			[0, 1, 0],  # up
			self.mvMatrix)

		self.texture_shader.use()
		self.texture_shader.mvMatrix = self.mvMatrix

		self.texture_shader.draw_mesh(self.ground_mesh)
		self.texture_shader.draw_mesh(self.sky_mesh)

		self.phong_shader.use()

		mat4.multiplyVec3(self.mvMatrix, self.lightLocation, self.transformedLightLocation)
		self.phong_shader.lightLocation = self.transformedLightLocation

		mat4.translate(self.mvMatrix, [-1.5, 0, -movement])
		self.robot.draw(self.mvMatrix)
		mat4.translate(self.mvMatrix, [1.5, 0, 0])
		self.robot.draw(self.mvMatrix)
		mat4.translate(self.mvMatrix, [1.5, 0, 0])
		self.robot.draw(self.mvMatrix)
		mat4.translate(self.mvMatrix, [-2.25, 0, 1])
		self.robokid.draw(self.mvMatrix)

		self.fader_layer.draw(beat)

		# sleep(0.01)
