from shortcrust.gl2 import *
from shortcrust.matrix import vec3, mat3, mat4
from shortcrust.buffer import AttributeBuffer
from shortcrust.geometry import Sphere
from shortcrust.shader import ShaderProgram

from shaders.phong import PhongTextureShader
from shaders.texture import TextureShader

import math


class TitleShader(ShaderProgram):
	vertex_shader = """
		attribute vec3 aVertexPosition;
		attribute vec2 aTexturePosition;

		varying vec2 vTextureCoord;

		void main(void) {
			gl_Position = vec4(aVertexPosition, 1.0);
			vTextureCoord = aTexturePosition;
		}
	"""

	fragment_shader = """
#ifdef GLES2
		precision mediump float;
#endif

		varying vec2 vTextureCoord;
		uniform sampler2D uSampler;
		uniform float uAlpha;

		void main(void) {
			gl_FragColor = vec4(
				texture2D(uSampler, vec2(vTextureCoord.s, vTextureCoord.t)).rgb,
				uAlpha);
		}
	"""

	def __init__(self):
		super(TitleShader, self).__init__()
		self.aVertexPosition = self.get_attrib('aVertexPosition')
		self.aTexturePosition = self.get_attrib('aTexturePosition')
		self.uSampler = self.get_uniform('uSampler')
		self.uAlpha = self.get_uniform('uAlpha')

	def draw_mesh(self, mesh, alpha):
		mesh.positions.attach(self.aVertexPosition)
		mesh.texture_coords.attach(self.aTexturePosition)
		glUniform1f(self.uAlpha, alpha)
		mesh.draw()


class TitleMesh(object):
	mode = GL_TRIANGLE_STRIP

	def __init__(self):
		self.positions = AttributeBuffer([
			(-1, -1, -0.999), (-1, 1, -0.999), (1, -1, -0.999), (1, 1, -0.999)
		])
		self.vertex_count = self.positions.element_count

		self.texture_coords = AttributeBuffer([
			(0.0, 0.75), (0.0, 0.0), (1.0, 0.75), (1, 0.0)
		])

	def draw(self):
		glDrawArrays(self.mode, 0, self.vertex_count)


class TitleLayer(object):
	def __init__(self, scene):
		self.shader = TitleShader()
		self.mesh = TitleMesh()
		self.texture = scene.app.textures['title']

	def draw(self, beat):
		if beat < 8:
			alpha = 1
		elif beat < 16:
			alpha = 1.0 - (beat - 8) / 8
		else:
			return
		self.shader.use()
		self.texture.activate(GL_TEXTURE0)
		glUniform1i(self.shader.uSampler, 0)
		self.shader.draw_mesh(self.mesh, alpha)


class FaderLayer(object):
	def __init__(self, shader, mesh):
		self.shader = shader
		self.mesh = mesh

	def draw(self, beat):
		if beat < 32:
			return
		else:
			alpha = (beat - 32) / 2
		self.shader.use()
		self.shader.draw_mesh(self.mesh, alpha)


class StarMesh(object):
	mode = GL_TRIANGLE_STRIP

	def __init__(self, app):
		self.positions = AttributeBuffer([
			(-10, -20, 10), (30, -20, 10), (-10, 10, 10), (30, 10, 10)
		])
		self.vertex_count = self.positions.element_count

		self.texture_positions = AttributeBuffer([
			(0.0, 0.0), (1.0, 0.0), (0.0, 1.0), (1.0, 1.0)
		])

		self.texture = app.textures['stars']

	def draw(self):
		glDrawArrays(self.mode, 0, self.vertex_count)


class PlanetScene(object):
	def __init__(self, app):
		glClearColor(0, 0, 0, 1.0)
		glEnable(GL_DEPTH_TEST)

		self.app = app
		self.phong_shader = PhongTextureShader()
		self.texture_shader = TextureShader()
		self.title_layer = TitleLayer(self)
		self.fader_layer = FaderLayer(app.fader_shader, app.fader_mesh)

		self.mvMatrix = mat4.create()
		self.pMatrix = mat4.create()
		self.nMatrix = mat3.create()
		# self.lightLocation = vec3.create([-3, 30, 0])
		self.transformedLightLocation = vec3.create()

		self.texture_shader.use()
		self.texture_shader.pMatrix = mat4.perspective(45, self.app.aspect_ratio, 0.1, 100.0)

		self.phong_shader.use()
		self.phong_shader.pMatrix = mat4.perspective(45, self.app.aspect_ratio, 0.1, 100.0)

		self.mesh = Sphere().to_mesh(texture=app.textures['planet'])

		self.star_mesh = StarMesh(app)

	def draw(self, beat):
		glClearColor(0, 0, 0, 1.0)
		glViewport(0, 0, self.app.width, self.app.height)
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

		orig_pos = [0, 0, -6]
		final_pos = [0, 0.5, -3]
		if beat < 20:
			progress = 0
		else:
			progress = (beat - 20) / 30.0
		pos = [
			orig_pos[0] * (1 - progress) + final_pos[0] * progress,
			orig_pos[1] * (1 - progress) + final_pos[1] * progress,
			orig_pos[2] * (1 - progress) + final_pos[2] * progress
		]

		mat4.lookAt(
			pos,  # eye
			[4 - beat / 12.0, 0, 0],  # lookat
			[0, 1, 0],  # up
			self.mvMatrix)

		self.texture_shader.use()
		self.texture_shader.mvMatrix = self.mvMatrix
		self.texture_shader.draw_mesh(self.star_mesh)

		self.phong_shader.use()
		self.phong_shader.mvMatrix = self.mvMatrix
		mat4.toInverseMat3(self.mvMatrix, self.nMatrix)
		mat3.transpose(self.nMatrix)
		self.phong_shader.nMatrix = self.nMatrix

		lightAngle = math.pi * 0.6 - beat / 14.0

		mat4.multiplyVec3(self.mvMatrix, [-1, 30 * math.cos(lightAngle), 30 * math.sin(lightAngle)], self.transformedLightLocation)
		self.phong_shader.lightLocation = self.transformedLightLocation

		# mat4.translate(self.mvMatrix, [-1.5, 0, 0])
		self.phong_shader.draw_mesh(self.mesh)

		self.title_layer.draw(beat)
		self.fader_layer.draw(beat)
