from shortcrust.geometry import Sphere, Cylinder, Union
from shortcrust.matrix import mat3, mat4
import math


BEER_ANIM = '................................................................................121212..12121212121212..12121212121332..12121212121332..1212133332....'
ROBO_ANIM = '......................................................................................11..............11..............11..............11..............'


class BaseRobot(object):
	def __init__(self, shader, color, torso_h=1.0):
		self.torso_h = torso_h
		self.shader = shader
		self.nMatrix = mat3.create()

		EYE_ANGLE = math.pi / 8.0

		self.robot_mesh = Union([
			# torso
			Cylinder(c=[0, 1, 0], r=0.5, h=torso_h),
			# torso top
			Sphere(min_lat_deg=0, c=[0, 1 + torso_h, 0], r=0.5, scale=[1, 0.25, 1], lat_divisions=4),
			# neck
			Cylinder(c=[0, 1.125 + torso_h, 0], r=0.125, h=0.125, divisions=4),
			# head
			Cylinder(c=[0, 1.375 + torso_h, 0], r=0.25, h=0.25),
			# scalp
			Sphere(min_lat_deg=0, c=[0, 1.625 + torso_h, 0], r=0.25, scale=[1, 0.6, 1], lat_divisions=4),
		]).to_mesh(material_color=color)

		self.robot_left_eye_mesh = Sphere(
			c=[-math.sin(EYE_ANGLE) * 0.25, 1.54 + torso_h, -math.cos(EYE_ANGLE) * 0.25], r=0.05,
			min_lng_deg=90, max_lng_deg=270, lng_divisions=4, lat_divisions=4
		).to_mesh(material_color=[0.05, 0.05, 0.05])
		self.robot_right_eye_mesh = Sphere(
			c=[math.sin(EYE_ANGLE) * 0.25, 1.54 + torso_h, -math.cos(EYE_ANGLE) * 0.25], r=0.05,
			min_lng_deg=90, max_lng_deg=270, lng_divisions=4, lat_divisions=4
		).to_mesh(material_color=[0.05, 0.05, 0.05])

		self.robot_left_leg_mesh = Cylinder(c=[-0.25, -1, 0], r=0.125, h=1).to_mesh(
			material_color=color
		)

		self.robot_right_leg_mesh = Cylinder(c=[+0.25, -1, 0], r=0.125, h=1).to_mesh(
			material_color=color
		)

		self.robot_jaw_mesh = Cylinder(c=[0, -0.0625, 0], r=0.26, h=0.125).to_mesh(
			material_color=color
		)

		self.robot_left_arm_mesh = Union([
			# left shoulder
			Sphere(min_lat_deg=0, c=[-0.6, 0, 0], r=0.1, scale=[1, 0.75, 1]),
			# left arm
			Cylinder(c=[-0.6, -0.9, 0], r=0.1, h=0.9),
			# left hand
			Sphere(max_lat_deg=0, c=[-0.6, -0.9, 0], r=0.1),
		]).to_mesh(material_color=color)
		self.robot_right_arm_mesh = Union([
			# right shoulder
			Sphere(min_lat_deg=0, c=[+0.6, 0, 0], r=0.1, scale=[1, 0.75, 1]),
			# right arm
			Cylinder(c=[+0.6, -0.9, 0], r=0.1, h=0.9),
			# right hand
			Sphere(max_lat_deg=0, c=[+0.6, -0.9, 0], r=0.1),
		]).to_mesh(material_color=color)

		self.body_mvMatrix = mat4.create()
		self.left_arm_mvMatrix = mat4.create()
		self.right_arm_mvMatrix = mat4.create()
		self.left_leg_mvMatrix = mat4.create()
		self.right_leg_mvMatrix = mat4.create()
		self.jaw_mvMatrix = mat4.create()

	def draw(self, mvMatrix):
		mat4.toInverseMat3(mvMatrix, self.nMatrix)
		mat3.transpose(self.nMatrix)
		self.shader.nMatrix = self.nMatrix

		mat4.translate(mvMatrix, [0, self.jump_height, 0], self.body_mvMatrix)
		self.shader.mvMatrix = self.body_mvMatrix

		self.shader.draw_mesh(self.robot_mesh)
		self.shader.draw_mesh(self.robot_left_eye_mesh)
		self.shader.draw_mesh(self.robot_right_eye_mesh)

		mat4.translate(self.body_mvMatrix, [0, 1.3125 + self.torso_h, 0], self.jaw_mvMatrix)
		mat4.rotateX(self.jaw_mvMatrix, self.jaw_angle)
		self.shader.mvMatrix = self.jaw_mvMatrix
		mat4.toInverseMat3(self.jaw_mvMatrix, self.nMatrix)
		mat3.transpose(self.nMatrix)
		self.shader.nMatrix = self.nMatrix
		self.shader.draw_mesh(self.robot_jaw_mesh)

		mat4.translate(self.body_mvMatrix, [0, 1, 0], self.left_leg_mvMatrix)
		mat4.rotateX(self.left_leg_mvMatrix, -self.arm_angle)
		self.shader.mvMatrix = self.left_leg_mvMatrix
		mat4.toInverseMat3(self.left_leg_mvMatrix, self.nMatrix)
		mat3.transpose(self.nMatrix)
		self.shader.nMatrix = self.nMatrix
		self.shader.draw_mesh(self.robot_left_leg_mesh)

		mat4.translate(self.body_mvMatrix, [0, 1, 0], self.right_leg_mvMatrix)
		mat4.rotateX(self.right_leg_mvMatrix, self.arm_angle)
		self.shader.mvMatrix = self.right_leg_mvMatrix
		mat4.toInverseMat3(self.right_leg_mvMatrix, self.nMatrix)
		mat3.transpose(self.nMatrix)
		self.shader.nMatrix = self.nMatrix
		self.shader.draw_mesh(self.robot_right_leg_mesh)

		mat4.translate(self.body_mvMatrix, [0, 0.9 + self.torso_h, 0], self.left_arm_mvMatrix)
		mat4.rotateX(self.left_arm_mvMatrix, self.arm_angle)
		self.shader.mvMatrix = self.left_arm_mvMatrix
		mat4.toInverseMat3(self.left_arm_mvMatrix, self.nMatrix)
		mat3.transpose(self.nMatrix)
		self.shader.nMatrix = self.nMatrix
		self.shader.draw_mesh(self.robot_left_arm_mesh)

		mat4.translate(self.body_mvMatrix, [0, 0.9 + self.torso_h, 0], self.right_arm_mvMatrix)
		mat4.rotateX(self.right_arm_mvMatrix, -self.arm_angle)
		self.shader.mvMatrix = self.right_arm_mvMatrix
		mat4.toInverseMat3(self.right_arm_mvMatrix, self.nMatrix)
		mat3.transpose(self.nMatrix)
		self.shader.nMatrix = self.nMatrix
		self.shader.draw_mesh(self.robot_right_arm_mesh)

	def animate(self, beat):
		self.arm_angle = math.sin(beat * math.pi) * math.pi / 6
		self.jump_height = 0


class Robot(BaseRobot):
	def __init__(self, shader):
		super(Robot, self).__init__(shader, [0.2, 0.3, 0.2])

	def animate(self, beat):
		super(Robot, self).animate(beat)

		mouth_beat = beat + 0.2  # offset so that mouth has chance to open before the beat
		try:
			beer_phase = BEER_ANIM[int(mouth_beat * 2)]
		except IndexError:
			beer_phase = '.'

		if beer_phase == '1' or beer_phase == '2':
			self.jaw_angle = -math.fabs(math.sin(mouth_beat * math.pi) * math.pi / 6)
		elif beer_phase == '3':
			self.jaw_angle = -math.pi / 6
		else:
			self.jaw_angle = 0


class Robokid(BaseRobot):
	def __init__(self, shader):
		super(Robokid, self).__init__(shader, [0.6, 0.2, 0.16], 0.7)

	def animate(self, beat):
		super(Robokid, self).animate(beat)

		mouth_beat = beat + 0.2  # offset so that mouth has chance to open before the beat

		try:
			robot_phase = ROBO_ANIM[int(mouth_beat * 2)]
		except IndexError:
			robot_phase = '.'

		if robot_phase == '1':
			self.jaw_angle = -math.fabs(math.sin(mouth_beat * 2 * math.pi) * math.pi / 6)
		else:
			self.jaw_angle = 0

		jump_beat = beat + 0.3
		try:
			jump_phase = ROBO_ANIM[int(jump_beat * 2)]
		except IndexError:
			jump_phase = '.'

		if jump_phase == '1':
			self.jump_height = math.fabs(math.sin(jump_beat * math.pi)) * 0.5
		else:
			self.jump_height = 0
