from shortcrust.shader import ShaderProgram
from shortcrust.gl2 import *


class PhongSolidColorShader(ShaderProgram):
	vertex_shader = """
		attribute vec3 aVertexPosition;
		attribute vec3 aVertexNormal;

		uniform mat4 uMVMatrix;
		uniform mat4 uPMatrix;
		uniform mat3 uNMatrix;

		varying vec3 vNormal;
		varying vec4 vPosition;

		void main(void) {
			/* apply MVMatrix to move vertex into correct position within the scene */
			vPosition = uMVMatrix * vec4(aVertexPosition, 1.0);
			/* perform equivalent transform on normal */
			vNormal = uNMatrix * aVertexNormal;

			/* project vertex into screen space */
			gl_Position = uPMatrix * vPosition;
		}
	"""

	fragment_shader = """
#ifdef GLES2
		precision mediump float;
#endif

		varying vec3 vNormal;
		varying vec4 vPosition;

		uniform vec3 uLightLocation;
		uniform vec3 uMaterialColor;

		const vec3 cAmbientColor = vec3(0.01, 0.01, 0.01);
		const vec3 cDiffuseColor = vec3(0.8, 0.8, 0.8);
		const vec3 cSpecularColor = vec3(0.3, 0.3, 0.3);
		const float cMaterialShininess = 3.0;

		void main(void) {
			/* calculate diffuse light intensity - angle between light and surface normal */
			vec3 lightDirection = normalize(uLightLocation - vPosition.xyz);
			float diffuseLightIntensity = max(dot(vNormal, lightDirection), 0.0);

			/* calculate specular light intensity - angle between eye and reflected light */
			vec3 eyeDirection = normalize(-vPosition.xyz);
			vec3 reflectionDirection = reflect(-lightDirection, vNormal);
			float specularLightIntensity = pow(max(dot(reflectionDirection, eyeDirection), 0.0), cMaterialShininess);

			/* add ambient, diffuse and specular components to give final light colour */
			vec3 lightWeighting = cAmbientColor + cDiffuseColor * diffuseLightIntensity + cSpecularColor + specularLightIntensity;

			gl_FragColor = vec4(uMaterialColor * lightWeighting, 1.0);
			/* gl_FragColor = vec4(lightDirection, 1.0); */
		}
	"""

	def __init__(self):
		super(PhongSolidColorShader, self).__init__()
		self.use()

		self.aVertexPosition = self.get_attrib('aVertexPosition')
		self.aVertexNormal = self.get_attrib('aVertexNormal')

		self.uPMatrix = self.get_uniform('uPMatrix')
		self.uMVMatrix = self.get_uniform('uMVMatrix')
		self.uNMatrix = self.get_uniform('uNMatrix')
		self.uLightLocation = self.get_uniform('uLightLocation')
		self.uMaterialColor = self.get_uniform('uMaterialColor')

	def _set_pMatrix(self, pMatrix):
		glUniformMatrix4fv(self.uPMatrix, 1, GL_FALSE, pMatrix)

	pMatrix = property(fset=_set_pMatrix)

	def _set_mvMatrix(self, mvMatrix):
		glUniformMatrix4fv(self.uMVMatrix, 1, GL_FALSE, mvMatrix)

	mvMatrix = property(fset=_set_mvMatrix)

	def _set_nMatrix(self, nMatrix):
		glUniformMatrix3fv(self.uNMatrix, 1, GL_FALSE, nMatrix)

	nMatrix = property(fset=_set_nMatrix)

	def _set_lightLocation(self, lightLocation):
		glUniform3fv(self.uLightLocation, 1, lightLocation)

	lightLocation = property(fset=_set_lightLocation)

	def _set_materialColor(self, materialColor):
		glUniform3fv(self.uMaterialColor, 1, materialColor)

	materialColor = property(fset=_set_materialColor)

	def draw_mesh(self, mesh):
		self.materialColor = mesh.material_color
		mesh.positions.attach(self.aVertexPosition)
		mesh.normals.attach(self.aVertexNormal)
		mesh.draw()


class PhongTextureShader(ShaderProgram):
	vertex_shader = """
		attribute vec3 aVertexPosition;
		attribute vec3 aVertexNormal;
		attribute vec2 aTexturePosition;

		uniform mat4 uMVMatrix;
		uniform mat4 uPMatrix;
		uniform mat3 uNMatrix;

		varying vec3 vNormal;
		varying vec4 vPosition;
		varying vec2 vTextureCoord;

		void main(void) {
			/* apply MVMatrix to move vertex into correct position within the scene */
			vPosition = uMVMatrix * vec4(aVertexPosition, 1.0);
			/* perform equivalent transform on normal */
			vNormal = uNMatrix * aVertexNormal;

			/* project vertex into screen space */
			gl_Position = uPMatrix * vPosition;

			/* pass texture coordinates through to fragment shader for interpolation */
			vTextureCoord = aTexturePosition;
		}
	"""

	fragment_shader = """
#ifdef GLES2
		precision mediump float;
#endif

		varying vec3 vNormal;
		varying vec4 vPosition;
		varying vec2 vTextureCoord;

		uniform vec3 uLightLocation;
		uniform sampler2D uSampler;

		const vec3 cAmbientColor = vec3(0.01, 0.01, 0.01);
		const vec3 cDiffuseColor = vec3(0.8, 0.8, 0.8);
		const vec3 cSpecularColor = vec3(0.0, 0.0, 0.0);
		const float cMaterialShininess = 3.0;

		void main(void) {
			/* calculate diffuse light intensity - angle between light and surface normal */
			vec3 lightDirection = normalize(uLightLocation - vPosition.xyz);
			float diffuseLightIntensity = max(dot(vNormal, lightDirection), 0.0);

			/* calculate specular light intensity - angle between eye and reflected light */
			vec3 eyeDirection = normalize(-vPosition.xyz);
			vec3 reflectionDirection = reflect(-lightDirection, vNormal);
			float specularLightIntensity = pow(max(dot(reflectionDirection, eyeDirection), 0.0), cMaterialShininess);

			/* add ambient, diffuse and specular components to give final light colour */
			vec3 lightWeighting = cAmbientColor + cDiffuseColor * diffuseLightIntensity + cSpecularColor + specularLightIntensity;

			vec3 materialColor = texture2D(uSampler, vec2(vTextureCoord.s, vTextureCoord.t)).rgb;
			/* vec3 materialColor = vec3(vTextureCoord.s, vTextureCoord.t, 0.5); */

			gl_FragColor = vec4(materialColor * lightWeighting, 1.0);
			/* gl_FragColor = vec4(lightDirection, 1.0); */
		}
	"""

	def __init__(self):
		super(PhongTextureShader, self).__init__()
		self.use()

		self.aVertexPosition = self.get_attrib('aVertexPosition')
		self.aVertexNormal = self.get_attrib('aVertexNormal')
		self.aTexturePosition = self.get_attrib('aTexturePosition')

		self.uPMatrix = self.get_uniform('uPMatrix')
		self.uMVMatrix = self.get_uniform('uMVMatrix')
		self.uNMatrix = self.get_uniform('uNMatrix')
		self.uLightLocation = self.get_uniform('uLightLocation')
		self.uSampler = self.get_uniform('uSampler')

	def _set_pMatrix(self, pMatrix):
		glUniformMatrix4fv(self.uPMatrix, 1, GL_FALSE, pMatrix)

	pMatrix = property(fset=_set_pMatrix)

	def _set_mvMatrix(self, mvMatrix):
		glUniformMatrix4fv(self.uMVMatrix, 1, GL_FALSE, mvMatrix)

	mvMatrix = property(fset=_set_mvMatrix)

	def _set_nMatrix(self, nMatrix):
		glUniformMatrix3fv(self.uNMatrix, 1, GL_FALSE, nMatrix)

	nMatrix = property(fset=_set_nMatrix)

	def _set_lightLocation(self, lightLocation):
		glUniform3fv(self.uLightLocation, 1, lightLocation)

	lightLocation = property(fset=_set_lightLocation)

	def draw_mesh(self, mesh):
		mesh.texture.activate(GL_TEXTURE0)
		glUniform1i(self.uSampler, 0)
		self.materialColor = mesh.material_color
		mesh.positions.attach(self.aVertexPosition)
		mesh.normals.attach(self.aVertexNormal)
		mesh.texture_positions.attach(self.aTexturePosition)
		mesh.draw()
